function p = Q_sweep(Ps,Rhc,qns,K,x0)
N = length(Rhc);
P = zeros(3,4,N);
A = zeros(2,3,N);
b = zeros(2,N);
c = zeros(N,3);
d = zeros(N,1);
for i = 1:N
    P(:,:,i) = K*[Rhc{i} -Rhc{i}*Ps(:,i)];
    A(:,:,i) = qns(:,i)*P(3,1:3,i)-P(1:2,1:3,i);
    b(:,i) = P(3,4,i)*qns(:,i)-P(1:2,4,i);
    c(i,:) = P(3,1:3,i);
    d(i) = P(3,4,i);
end
[A4,b4] = four(A,b);
x = x0;
for itr = 1:50
    dx = DESCENTDIR(A4,b4,c,d,x);
    dx = dx/norm(dx);
    u = zeros(N,4);
    v = zeros(N,4);
    w = zeros(N,1);
    z = zeros(N,1);
    for i = 1:N
        for j = 1:4
            u(i,j) = A4(j,:,i)*dx;
            v(i,j) = A4(j,:,i)*x+b4(j,i);
            w(i) = c(i,:)*dx;
            z(i) = c(i,:)*x+d(i);
        end
    end
    am = alphaMax(z,w);
    e_E = Extremity(u,v,w,z);
    e_I = Intersection(u,v,w,z);
    a = sort([e_E.alpha e_I.alpha am]);
    a = a(a<am);
    f = zeros(1,length(a));
    for k = 1:length(a)
        r = zeros(N,4);
        for i = 1:N
            for j = 1:4
                r(i,j) = (u(i,j)*a(k)+v(i,j))/(w(i)*a(k)+z(i));
            end
        end
        ri = max(r,[],2);
        f(k) = median(ri);
    end
    [~,i] = min(f);
    if isempty(a)
        break;
    elseif abs(a(i)) < 1e-14 && itr > 5
        break;
    end
    x = x+dx*a(i);
end
% disp(itr)
p = x;

function am = alphaMax(z,w)
zw = -z./w;
am = min(zw(zw>0));
if isempty(am)
    am = inf;
end

function e = Extremity(u,v,w,z)
N = length(z);
c = combnk(1:4,2);
e_alpha = zeros(1,N);
e_i = zeros(1,N);
for i = 1:N
    rt = zeros(1,6);
    vrt = zeros(1,6);
    for k = 1:size(c,1)
        j1 = c(k,1);
        j2 = c(k,2);
        rt(k) = (v(i,j2)-v(i,j1))/(u(i,j1)-u(i,j2));
        vrt(k) = (u(i,j1)*rt(k)+v(i,j1))./(w(i)*rt(k)+z(i));
    end
    vrt = vrt(rt>0);
    rt = rt(rt>0);
    [vrt,I] = sort(vrt,'ascend');
    rt = rt(I);
    r = [];
    r(1,:) = (u(i,1)*rt+v(i,1))./(w(i)*rt+z(i));
    r(2,:) = (u(i,2)*rt+v(i,2))./(w(i)*rt+z(i));
    r(3,:) = (u(i,3)*rt+v(i,3))./(w(i)*rt+z(i));
    r(4,:) = (u(i,4)*rt+v(i,4))./(w(i)*rt+z(i));
    for k = 1:length(rt)
        if vrt(k) == max(r(:,k))
            e_alpha(i) = rt(k);
            e_i(i) = i;
            break;
        end
    end
end
e.alpha = e_alpha(e_alpha>0);
e.i1 = e_i(e_alpha>0);

function e = Intersection(u,v,w,z)
N = length(z);
c = combnk(1:N,2);
e_alpha = zeros(1,1);
e_i1 = zeros(1,1);
e_i2 = zeros(1,1);
n = 0;
for i = 1:size(c,1)
    i1 = c(i,1);
    i2 = c(i,2);
    for j1 = 1:4
        for j2 = 1:4
            p2 = u(i1,j1)*w(i2)-u(i2,j2)*w(i1);
            p1 = u(i1,j1)*z(i2)+v(i1,j1)*w(i2)-u(i2,j2)*z(i1)-v(i2,j2)*w(i1);
            p0 = v(i1,j1)*z(i2)-v(i2,j2)*z(i1);
            rt = roots([p2 p1 p0]);
            for k = 1:length(rt)
                if rt(k)<=0 || ~isreal(rt(k))
                    continue
                end
                vrt = (u(i1,j1)*rt(k)+v(i1,j1))/(w(i1)*rt(k)+z(i1));
                r1(1) = (u(i1,1)*rt(k)+v(i1,1))/(w(i1)*rt(k)+z(i1));
                r1(2) = (u(i1,2)*rt(k)+v(i1,2))/(w(i1)*rt(k)+z(i1));
                r1(3) = (u(i1,3)*rt(k)+v(i1,3))/(w(i1)*rt(k)+z(i1));
                r1(4) = (u(i1,4)*rt(k)+v(i1,4))/(w(i1)*rt(k)+z(i1));
                r2(1) = (u(i2,1)*rt(k)+v(i2,1))/(w(i2)*rt(k)+z(i2));
                r2(2) = (u(i2,2)*rt(k)+v(i2,2))/(w(i2)*rt(k)+z(i2));
                r2(3) = (u(i2,3)*rt(k)+v(i2,3))/(w(i2)*rt(k)+z(i2));
                r2(4) = (u(i2,4)*rt(k)+v(i2,4))/(w(i2)*rt(k)+z(i2));
                if abs(vrt-max(r1))<1e-12 && abs(vrt-max(r2))<1e-12
                    n = n + 1;
                    e_alpha(n) = rt(k);
                    e_i1(n) = i1;
                    e_i2(n) = i2;
                end
            end
        end
    end
end
e.alpha = e_alpha;
e.i1 = e_i1;
e.i2 = e_i2;

function [A4,b4] = four(A,b)
N = size(b,2);
A4 = zeros(4,3,N);
b4 = zeros(4,N);
for i = 1:N
    A4(1,:,i) = +A(1,:,i);
    A4(2,:,i) = -A(1,:,i);
    A4(3,:,i) = +A(2,:,i);
    A4(4,:,i) = -A(2,:,i);
    b4(1,i) = +b(1,i);
    b4(2,i) = -b(1,i);
    b4(3,i) = +b(2,i);
    b4(4,i) = -b(2,i);
end

function dx = DESCENTDIR(A4,b4,c,d,x)
N = length(d);
r = zeros(N,4);
for i = 1:N
    for j = 1:4
        r(i,j) = (A4(j,:,i)*x+b4(j,i))/(c(i,:)*x+d(i));
    end
end
ri = max(r,[],2);
if mod(N,2)==1
    rh = median(ri);
else
    rh = median([ri;inf]);
end
ActiveSet = (ri == rh);
idx = find(ActiveSet);
Na = sum(ActiveSet);
J = zeros(1,2);
n = zeros(3,1);
k = 0;
for i = 1:Na
    for j = 1:4
        if r(idx(i),j)==rh
            k = k + 1;
            J(k,:) = [idx(i),j];
            n(:,k) = (-A4(j,:,idx(i))+rh*c(idx(i),:))';
        end
    end
end
dx = [0;0;0];
if k == 1
    dx = n(:,1);
elseif k == 2
    dx = n(:,1) + n(:,2);
else
    c = combnk(1:k,3);
    for i = 1:size(c,1)
        y = 0;
        y = y + cross(n(:,c(i,1)),n(:,c(i,2)));
        y = y + cross(n(:,c(i,2)),n(:,c(i,3)));
        y = y + cross(n(:,c(i,3)),n(:,c(i,1)));
        y = y*sign(y'*n(:,c(i,1)));
        for j = 1:k
            if y'*n(:,j)<0
                continue;
            end
        end
        dx = y;
        break;
    end
end
